from typing import List

import plotly.graph_objects as go

from ..beam_class import Beam
from ..constants import (
    COLOR_WIRE_FRAME_BEAM,
    COLOR_WIRE_FRAME_DETECTOR,
    COLOR_WIRE_FRAME_PAD,
    COLOR_WIRE_FRAME_TABLE,
    PLOT_TRACE_ORDER_BEAM_WIREFRAME,
    PLOT_TRACE_ORDER_DETECTOR_WIREFRAME,
    PLOT_TRACE_ORDER_PHANTOM_WIREFRAME,
)
from ..phantom_class import Phantom


def create_wireframes(beam: Beam, table: Phantom, pad: Phantom, line_width: int = 4, visible: bool = True):
    """Create wireframe versions of the mesh3d objects in plot_geometry.

    The purpose of this function is to enhance the plot_geometry plot by adding
    wireframes.

    Parameters
    ----------
    beam: Beam
        X-ray beam, i.e. instance of class Beam. See beam_class for more info
    table: Phantom
        Table phantom from instance of class Phantom. phantom_type must be "table"
    pad: Phantom
        Table phantom from instance of class Phantom. phantom_type must be "pad"
    line_width : int, optional
        Line width of the wireframes. Default value is 4.
    visible : bool, optional
        Set the visibility of each of the wireframe traces

    """
    wf_beam = _create_beam_wireframe(beam=beam, line_width=line_width, visible=visible)

    wf_table = _create_phantom_object_wireframe(
        obj=table, color=COLOR_WIRE_FRAME_TABLE, line_width=line_width, visible=visible
    )

    wf_pad = _create_phantom_object_wireframe(
        obj=pad, color=COLOR_WIRE_FRAME_PAD, line_width=line_width, visible=visible
    )

    wf_detector = _create_detector_wire_frame(beam=beam, line_width=line_width, visible=visible)

    return wf_beam, wf_table, wf_pad, wf_detector


def _create_beam_wireframe(beam: Beam, line_width: int, visible: bool) -> go.Scatter3d:
    # This funciton creates a wireframe plot for the X-ray beam

    x = beam.r[:, 0].tolist()
    y = beam.r[:, 1].tolist()
    z = beam.r[:, 2].tolist()

    temp_x = []
    temp_y = []
    temp_z = []

    # connect trace by using a plot trace order (trace out path of wireframe by index of each point)
    for trace in PLOT_TRACE_ORDER_BEAM_WIREFRAME:
        temp_x.append(x[trace])
        temp_y.append(y[trace])
        temp_z.append(z[trace])

    return _create_wireframe_scatter3d(
        x=temp_x,
        y=temp_y,
        z=temp_z,
        line_width=line_width,
        visible=visible,
        color=COLOR_WIRE_FRAME_BEAM,
    )


def _create_phantom_object_wireframe(obj: Phantom, color: str, line_width: int, visible: bool) -> go.Scatter3d:
    # This function creates a wireframe plot for the X-ray table or pad

    x = obj.r[:, 0].tolist()
    y = obj.r[:, 1].tolist()
    z = obj.r[:, 2].tolist()

    temp_x = []
    temp_y = []
    temp_z = []

    # connect trace by using a plot trace order (trace out path of wireframe by index of each point)
    for trace in PLOT_TRACE_ORDER_PHANTOM_WIREFRAME:
        temp_x.append(x[trace])
        temp_y.append(y[trace])
        temp_z.append(z[trace])

    return _create_wireframe_scatter3d(
        x=temp_x,
        y=temp_y,
        z=temp_z,
        line_width=line_width,
        visible=visible,
        color=color,
    )


def _create_detector_wire_frame(beam: Beam, line_width: int, visible: bool) -> go.Scatter3d:
    # This funciton creates a wireframe plot for the X-ray detector

    x = beam.det_r[:, 0].tolist()
    y = beam.det_r[:, 1].tolist()
    z = beam.det_r[:, 2].tolist()

    temp_x = []
    temp_y = []
    temp_z = []

    # connect trace by using a plot trace order (trace out path of wireframe by index of each point)
    for trace in PLOT_TRACE_ORDER_DETECTOR_WIREFRAME:
        temp_x.append(x[trace])
        temp_y.append(y[trace])
        temp_z.append(z[trace])

    return _create_wireframe_scatter3d(
        x=temp_x,
        y=temp_y,
        z=temp_z,
        line_width=line_width,
        visible=visible,
        color=COLOR_WIRE_FRAME_DETECTOR,
    )


def _create_wireframe_scatter3d(
    x: List[float],
    y: List[float],
    z: List[float],
    line_width: int,
    visible: bool,
    color: str,
) -> go.Scatter3d:
    return go.Scatter3d(
        x=x,
        y=y,
        z=z,
        mode="lines",
        hoverinfo="skip",
        visible=visible,
        line=dict(color=color, width=line_width),
    )



import logging
from typing import Any, Dict, Optional, Tuple

import numpy as np
import pandas as pd

from pyskindose import constants as c
from pyskindose.calculate_dose.calculate_irradiation_event_result import (
    calculate_irradiation_event_result,
)
from pyskindose.corrections import calculate_k_bs, calculate_k_tab
from pyskindose.geom_calc import (
    check_new_geometry,
    fetch_and_append_hvl,
    position_patient_phantom_on_table,
)
from pyskindose.phantom_class import Phantom
from pyskindose.settings import PyskindoseSettings

logger = logging.getLogger(__name__)


def calculate_dose(
    normalized_data: pd.DataFrame,
    settings: PyskindoseSettings,
    table: Phantom,
    pad: Phantom,
) -> Tuple[Phantom, Optional[Dict[str, Any]]]:
    """Calculate skin dose.

    This function initializes the skin dose calculations.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    settings : PyskindoseSettings
        Settings class for PySkinDose
    table : Phantom
        Patient support table phantom
    pad : Phantom
        Patient support pad phantom

    Returns
    -------
    Tuple[Phantom, Optional[Dict[str, Any]]]
        [description]

    """
    if settings.mode != c.MODE_CALCULATE_DOSE and settings.mode != c.MODE_PLOT_DOSEMAP:
        logger.debug("Mode not set to calculate dose. Returning without doing anything")
        return None, None

    logger.info("Start performing dose calculations")
    patient = Phantom(
        phantom_model=settings.phantom.model,
        phantom_dim=settings.phantom.dimension,
        human_mesh=settings.phantom.human_mesh,
    )

    # position objects in starting position
    position_patient_phantom_on_table(
        patient=patient,
        table=table,
        pad=pad,
        pad_thickness=settings.phantom.dimension.pad_thickness,
        patient_offset=[
            settings.phantom.patient_offset.d_lon,
            settings.phantom.patient_offset.d_ver,
            settings.phantom.patient_offset.d_lat,
        ],
        patient_orientation=settings.phantom.patient_orientation,
    )

    normalized_data = fetch_and_append_hvl(
        data_norm=normalized_data,
        inherent_filtration=settings.inherent_filtration,
        corrections_db=settings.corrections_db_path,
    )

    # Check which irradiation events that contains updated
    # geometry parameters since the previous irradiation event
    new_geometry = check_new_geometry(normalized_data)

    # fetch of k_bs interpolation object (k_bs=f(field_size))for all events
    back_scatter_interpolation = calculate_k_bs(data_norm=normalized_data)

    k_tab = calculate_k_tab(
        data_norm=normalized_data,
        estimate_k_tab=settings.estimate_k_tab,
        k_tab_val=settings.k_tab_val,
        corrections_db=settings.corrections_db_path,
    )

    total_number_of_events = len(normalized_data)

    if settings.plot.notebook_mode:
        from tqdm import tqdm_notebook as pbar
    else:
        from tqdm import tqdm as pbar

    output_template = {
        c.OUTPUT_KEY_HITS: [[]] * total_number_of_events,
        c.OUTPUT_KEY_KERMA: [np.array] * total_number_of_events,
        c.OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW: [[]] * total_number_of_events,
        c.OUTPUT_KEY_CORRECTION_BACK_SCATTER: [[]] * total_number_of_events,
        c.OUTPUT_KEY_CORRECTION_MEDIUM: [[]] * total_number_of_events,
        c.OUTPUT_KEY_CORRECTION_TABLE: [[]] * total_number_of_events,
        c.OUTPUT_KEY_DOSE_MAP: np.zeros(len(patient.r)),
    }

    output = calculate_irradiation_event_result(
        normalized_data=normalized_data,
        event=0,
        total_events=len(normalized_data),
        new_geometry=new_geometry,
        k_tab=k_tab,
        hits=[],
        patient=patient,
        table=table,
        pad=pad,
        back_scatter_interpolation=back_scatter_interpolation,
        output=output_template,
        pbar=pbar(total=total_number_of_events, leave=False, desc="calculating skindose"),
        corrections_db=settings.corrections_db_path,
    )

    return patient, output



import logging

import pandas as pd

from ..beam_class import Beam
from ..constants import MODE_PLOT_SETUP
from ..phantom_class import Phantom
from .create_geometry_plot_texts import create_geometry_plot_texts
from .create_setup_and_event_plot import create_setup_and_event_plot

logger = logging.getLogger(__name__)


def plot_setup(
    mode: str,
    data_norm: pd.DataFrame,
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    dark_mode: bool = True,
    notebook_mode: bool = False,
):
    """Debugging feature for visualizing the geometry setup.

    This function plots the patient, table and pad in reference position together with
    the X-ray system with zero angulation.

    Parameters
    ----------
    mode : str
        The function will only run if this is set to "plot_setup".
    data_norm : pd.DataFrame
        Table containing dicom RDSR information from each irradiation event. See
        rdsr_normalizer.py for more information.
    patient : Phantom
        Patient phantom from instance of class Phantom. Can be of phantom_model "plane",
        "cylinder" or "human"
    table : Phantom
        Table phantom from instance of class Phantom. phantom_model must be "table"
    pad : Phantom
        Pad phantom from instance of class Phantom. phantom_model must be "pad"
    dark_mode : bool, optional
        set dark mode for plot
    notebook_mode : bool, optional
        optimize plot size for notebooks, default is True.

    """
    if mode != MODE_PLOT_SETUP:
        return

    logger.info("plotting geometry setup...")

    logger.debug("Creating beam")
    beam = Beam(data_norm, event=0, plot_setup=True)

    (
        source_text,
        beam_text,
        detectors_text,
        table_text,
        pad_text,
        patient_text,
    ) = create_geometry_plot_texts(beam=beam, table=table, pad=pad, patient=patient)

    # Define plot title
    title = f"<b>P</b>y<b>S</b>kin<b>D</b>ose [mode: {mode}]"

    create_setup_and_event_plot(
        mode=mode,
        title=title,
        patient=patient,
        patient_text=patient_text,
        table=table,
        table_text=table_text,
        pad=pad,
        pad_text=pad_text,
        beam=beam,
        beam_text=beam_text,
        source_text=source_text,
        detectors_text=detectors_text,
        dark_mode=dark_mode,
        notebook_mode=notebook_mode,
    )



import logging
from pathlib import Path
from typing import List

import numpy as np
import plotly.graph_objects as go
from PIL import Image

from pyskindose.phantom_class import Phantom
from pyskindose.plotting.create_layout_for_dose_map_plots import (
    create_layout_for_dose_map_plots,
)
from pyskindose.plotting.create_notebook_dose_map_plot import (
    create_notebook_dose_map_plot,
)
from pyskindose.settings import PyskindoseSettings

from ..constants import (
    DOSEMAP_COLORSCALE,
    PHANTOM_MODEL_PLANE,
    PLOT_EYE_BACK,
    PLOT_EYE_FRONT,
    PLOT_EYE_LEFT,
    PLOT_EYE_RIGHT,
    PLOT_FILE_TYPE_STATIC,
    PLOT_FONT_FAMILY,
    PLOT_ORDER_STATIC,
)
from .plot_settings import fetch_plot_colors, fetch_plot_margin, fetch_plot_size

logger = logging.getLogger(__name__)


def create_dose_map_plot(patient: Phantom, settings: PyskindoseSettings, dose_map: np.ndarray) -> None:
    """Plot a map of the absorbed skindose upon the patient phantom.

    This function creates and plots an offline plotly graph of the skin dose
    distribution on the phantom. The colorscale is mapped to the absorbed skin dose
    value. Only available for phantom type: "plane", "cylinder" or "human."

    Parameters
    ----------
    patient : Phantom
        patient phantom
    settings : PyskindoseSettings
        Settings class for PySkinDose
    dose_map : np.ndarray
        array with dose matrix, where each element is to be mapped to the corresponding
        skin cell of the patient.

    """
    if not settings.plot.plot_dosemap:
        logger.debug("Mode not set to plot dosemap. Returning without doing anything")
        return

    if dose_map is None:
        logger.debug(
            """Cannot plot dosemap since dose calculation has not been conducted.
            Returning without doing anything."""
        )
        raise ValueError(
            """Dosemap is None. Mode must be set to calculate dose in order to plot
            dosemap"""
        )

    # Fix error with plotly layout for 2D plane patient.
    if patient.phantom_model == PHANTOM_MODEL_PLANE:
        patient = Phantom(phantom_model=settings.phantom.model, phantom_dim=settings.phantom.dimension)

    # append dosemap to patient
    patient.dose = dose_map

    COLOR_CANVAS, COLOR_PLOT_TEXT, COLOR_GRID, COLOR_ZERO_LINE = fetch_plot_colors(dark_mode=settings.plot.dark_mode)

    PLOT_HEIGHT, PLOT_WIDTH = fetch_plot_size(notebook_mode=settings.plot.notebook_mode)

    PLOT_MARGINS = fetch_plot_margin(notebook_mode=settings.plot.notebook_mode)

    lat_text = [(f"<b>lat:</b> {np.around(patient.r[ind, 2],2)} cm<br>") for ind in range(len(patient.r))]

    lon_text = [f"<b>lon:</b> {np.around(patient.r[ind, 0],2)} cm<br>" for ind in range(len(patient.r))]

    ver_text = [f"<b>ver:</b> {np.around(patient.r[ind, 1],2)} cm<br>" for ind in range(len(patient.r))]

    dose_text = [f"<b>skin dose:</b> {round(patient.dose[ind],2)} mGy" for ind in range(len(patient.r))]

    hover_text = [lat_text[cell] + lon_text[cell] + ver_text[cell] + dose_text[cell] for cell in range(len(patient.r))]

    # create mesh object for the phantom
    phantom_mesh = [
        go.Mesh3d(
            x=patient.r[:, 0],
            y=patient.r[:, 1],
            z=patient.r[:, 2],
            i=patient.ijk[:, 0],
            j=patient.ijk[:, 1],
            k=patient.ijk[:, 2],
            intensity=patient.dose,
            colorscale=DOSEMAP_COLORSCALE,
            showscale=True,
            hoverinfo="text",
            text=hover_text,
            name="Human",
            colorbar=dict(
                tickfont=dict(color=COLOR_PLOT_TEXT),
                title=dict(
                    text="Skin dose [mGy]",
                    font=dict(family=PLOT_FONT_FAMILY, color=COLOR_PLOT_TEXT),
                ),
            ),
        )
    ]

    layout = create_layout_for_dose_map_plots(
        PLOT_MARGINS=PLOT_MARGINS,
        PLOT_HEIGHT=PLOT_HEIGHT,
        PLOT_WIDTH=PLOT_WIDTH,
        COLOR_PLOT_TEXT=COLOR_PLOT_TEXT,
        COLOR_CANVAS=COLOR_CANVAS,
    )

    # create figure
    fig = go.Figure(data=phantom_mesh, layout=layout)

    if settings.plot.interactivity:
        fig.show()
        return

    # proceed with creating static dose map
    eyes = [PLOT_EYE_RIGHT, PLOT_EYE_BACK, PLOT_EYE_LEFT, PLOT_EYE_FRONT]

    names: List[Path] = [
        settings.file_result_output_path / f"{filename}{PLOT_FILE_TYPE_STATIC}" for filename in PLOT_ORDER_STATIC
    ]

    if settings.plot.notebook_mode:
        from tqdm import tqdm_notebook as pbar

    else:
        from tqdm import tqdm as pbar

    # create static dose map plots
    try:
        for i in pbar(range(len(eyes)), desc="saving static dosemaps: "):
            fig["layout"]["scene"]["camera"] = eyes[i]
            fig.write_image(names[i].absolute())
    except IOError:
        logger.error("Failed to save image plots", exc_info=True)

    print(f"Saved plots to {settings.file_result_output_path.absolute()}")

    # show dose map plot with PIL if not in notebook mode
    if not settings.plot.notebook_mode:
        for image_file_name in names:
            im = Image.open(image_file_name.absolute())
            im.show()
        return

    # proceed with showing the dose map plot in notebook mode
    create_notebook_dose_map_plot(names=names)



import argparse
import logging
import sys
from pathlib import Path
from typing import Optional, Union

import pandas as pd

from pyskindose.analyze_data import analyze_data
from pyskindose.constants import (
    RUN_ARGUMENTS_MODE_GUI,
    RUN_ARGUMENTS_MODE_HEADLESS,
    RUN_ARGUMENTS_OUTPUT_DICT,
    RUN_ARGUMENTS_OUTPUT_HTML,
    RUN_ARGUMENTS_OUTPUT_JSON,
)
from pyskindose.dev_data import DEVELOPMENT_PARAMETERS
from pyskindose.helpers.parse_settings_to_settings_class import (
    parse_settings_to_settings_class,
)
from pyskindose.helpers.read_and_normalize_rdsr_data import read_and_normalise_rdsr_data
from pyskindose.settings import PyskindoseSettings

logger = logging.getLogger(__name__)


def main(file_path: Optional[str] = None, settings: Union[str, dict, PyskindoseSettings] = None):
    """Run PySkinDose.

    Copy settings_examples.json and save it as settings.json.
    Set all you parameters in this file. For debugging and developement,
    the PARAM_dev settings dictionary can be used by calling
    main(settings=PARAM_DEV).

    See settings.py for a description of all the parameters. Please visit
    https://github.com/rvbCMTS/PySkinDose for info on how to run
    PySkinDose.

    Parameters
    ----------
    file_path : str, optional
        file path to RDSR file or preparsed RDSR data in .json format
    settings : Union[str, dict, PyskindoseSettings], optional
        Setting file in either dict, json string format, or as a
        PyskindoseSettings object. By default, settings_examples.json is
        enabled.

    """
    settings = parse_settings_to_settings_class(settings=settings)

    data_norm = read_and_normalise_rdsr_data(rdsr_filepath=file_path, settings=settings)

    output = analyze_data(normalized_data=data_norm, settings=settings)

    if settings.output_format in ("dict", "json"):
        return output

def analyze_normalized_data_with_custom_settings_object(
    data_norm: pd.DataFrame,
    settings: Union[PyskindoseSettings, str, dict],
    output_format: Optional[str] = RUN_ARGUMENTS_OUTPUT_JSON,
) -> Union[str, dict]:
    """Run PySkinDose with custom normalized data and a custom specified settings objects.

    See the

    Parameters
    ----------
    data_norm : pd.DataFrame
        A pandas DataFrame containing the normalized data
    settings : Union[PySkinDoseSettings, str, dict]
        The settings for the PySkinDose analysis given as a PySKinDoseSettings object, a json-formatted string or a dict
    output_format : str, optional
        String specifying the output format. Valid values are "json"(default) and "dict"
    """
    if not isinstance(settings, (PyskindoseSettings, str, dict)):
        raise TypeError("Invalid type for input settings")

    if not isinstance(output_format, str) or not output_format.casefold() in [
        RUN_ARGUMENTS_OUTPUT_JSON,
        RUN_ARGUMENTS_OUTPUT_DICT,
        RUN_ARGUMENTS_OUTPUT_HTML,
    ]:
        raise ValueError(
            f"The output_format must be specified as a string with one of the valid values {RUN_ARGUMENTS_OUTPUT_JSON} "
            f"or {RUN_ARGUMENTS_OUTPUT_DICT}"
        )

    settings = parse_settings_to_settings_class(settings=settings)
    settings.output_format = output_format.casefold()

    return analyze_data(normalized_data=data_norm, settings=settings)


def get_argument_parser(arguments) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="PySkinDose",
        description=(
            "PySkinDose is a Python version 3.8 based program for patient peak"
            " skin dose (PSD) estimations from fluoroscopic procedures in"
            " interventional radiology."
        ),
    )
    parser.add_argument(
        "--mode",
        "-m",
        dest="mode",
        choices=(RUN_ARGUMENTS_MODE_HEADLESS, RUN_ARGUMENTS_MODE_GUI),
        default=RUN_ARGUMENTS_MODE_HEADLESS,
    )

    parser.add_argument(
        "--file-path",
        "-f",
        required=False,
        dest="file_path",
        help="Path to RDSR DICOM file (required in headless mode)",
    )

    parser.add_argument(
        "--settings",
        "-s",
        required=False,
        type=Path,
        default=None,
        dest="settings",
        help="Path to the settings file to use if a specific settings file is required",
    )

    return parser.parse_args(arguments)


if __name__ == "__main__":
    args = get_argument_parser(sys.argv)

    if (run_settings := args.settings) is None:
        logger.warning("No settings specified. Running with development parameters")
        run_settings = DEVELOPMENT_PARAMETERS

    main(file_path=args.file_path, settings=run_settings)



import logging

import plotly.graph_objects as go

from ..constants import (
    PLOT_ASPECTMODE_PLOT_DOSEMAP,
    PLOT_FONT_FAMILY,
    PLOT_FONT_SIZE,
    PLOT_HOVERLABEL_FONT_FAMILY,
    PLOT_HOVERLABEL_FONT_SIZE,
)

logger = logging.getLogger(__name__)


def create_layout_for_dose_map_plots(
    PLOT_MARGINS,
    PLOT_HEIGHT,
    PLOT_WIDTH,
    COLOR_PLOT_TEXT,
    COLOR_CANVAS,
):

    layout = go.Layout(
        height=PLOT_HEIGHT,
        width=PLOT_WIDTH,
        margin=PLOT_MARGINS,
        font=dict(family=PLOT_FONT_FAMILY, color=COLOR_PLOT_TEXT, size=PLOT_FONT_SIZE),
        hoverlabel=dict(font=dict(family=PLOT_HOVERLABEL_FONT_FAMILY, size=PLOT_HOVERLABEL_FONT_SIZE)),
        title=dict(font=dict(family=PLOT_FONT_FAMILY, size=PLOT_FONT_SIZE, color=COLOR_PLOT_TEXT)),
        paper_bgcolor=COLOR_CANVAS,
        scene=dict(
            aspectmode=PLOT_ASPECTMODE_PLOT_DOSEMAP,
            xaxis=dict(
                title="",
                backgroundcolor=COLOR_CANVAS,
                showgrid=False,
                zeroline=False,
                showticklabels=False,
            ),
            yaxis=dict(
                title="",
                backgroundcolor=COLOR_CANVAS,
                showgrid=False,
                zeroline=False,
                showticklabels=False,
            ),
            zaxis=dict(
                title="",
                backgroundcolor=COLOR_CANVAS,
                showgrid=False,
                zeroline=False,
                showticklabels=False,
            ),
        ),
    )

    return layout



from ..constants import (
    COLOR_CANVAS_DARK,
    COLOR_CANVAS_LIGHT,
    COLOR_GRID_DARK,
    COLOR_GRID_LIGHT,
    COLOR_PLOT_TEXT_DARK,
    COLOR_PLOT_TEXT_LIGHT,
    COLOR_SLIDER_BORDER_DARK,
    COLOR_SLIDER_BORDER_LIGHT,
    COLOR_SLIDER_TICK_DARK,
    COLOR_SLIDER_TICK_LIGHT,
    COLOR_ZERO_LINE_DARK,
    COLOR_ZERO_LINE_LIGHT,
    PLOT_HEIGHT,
    PLOT_HEIGHT_NOTEBOOK,
    PLOT_MARGIN,
    PLOT_MARGIN_NOTEBOOK,
    PLOT_SLIDER_PADDING,
    PLOT_SLIDER_PADDING_NOTEBOOK,
    PLOT_WIDTH,
    PLOT_WIDTH_NOTEBOOK,
)


def fetch_plot_colors(dark_mode: bool):
    """Fetch correct colors scheme for plotly plots.

    Parameters
    ----------
    dark_mode : bool
        Specifies whether dark mode should be implemented

    """
    if dark_mode:
        return (
            COLOR_CANVAS_DARK,
            COLOR_PLOT_TEXT_DARK,
            COLOR_GRID_DARK,
            COLOR_ZERO_LINE_DARK,
        )

    return (
        COLOR_CANVAS_LIGHT,
        COLOR_PLOT_TEXT_LIGHT,
        COLOR_GRID_LIGHT,
        COLOR_ZERO_LINE_LIGHT,
    )


def fetch_slider_colors(dark_mode: bool):
    """Fetch correct colors scheme for plotly sliders.

    Parameters
    ----------
    dark_mode : bool
        Specifies whether dark mode should be implemented

    """
    if dark_mode:
        return (
            COLOR_PLOT_TEXT_DARK,
            COLOR_SLIDER_TICK_DARK,
            COLOR_SLIDER_BORDER_DARK,
        )

    return COLOR_PLOT_TEXT_LIGHT, COLOR_SLIDER_TICK_LIGHT, COLOR_SLIDER_BORDER_LIGHT


def fetch_plot_size(notebook_mode: bool):
    """Fetch correct plot size for plotly plots.

    Parameters
    ----------
    notebook_mode : bool
        specifies whether plot size should be optimized for notebooks

    """
    if notebook_mode:
        return PLOT_HEIGHT_NOTEBOOK, PLOT_WIDTH_NOTEBOOK

    return PLOT_HEIGHT, PLOT_WIDTH


def fetch_plot_margin(notebook_mode: bool):
    """Fetch correct margins for plotly plots.

    Parameters
    ----------
    notebook_mode : bool
        Specifies whether plot margins should be optimized for notebooks

    """
    if notebook_mode:
        return PLOT_MARGIN_NOTEBOOK

    return PLOT_MARGIN


def fetch_slider_padding(notebook_mode: bool):
    """Fetch correct slider margins for plot procedure.

    Parameters
    ----------
    notebook_mode : bool
        Specifies whether the slider padding should be optimized for notebooks

    """
    if notebook_mode:
        return PLOT_SLIDER_PADDING_NOTEBOOK

    return PLOT_SLIDER_PADDING



from typing import Union

from ..beam_class import Beam
from ..constants import PHANTOM_MODEL_PLANE
from ..phantom_class import Phantom


def _get_visual_offset(patient: Union[Phantom, Beam]) -> float:
    """Set visual offset of phantom objects.

    Determines the visual offset needed for visualizing the phantom correctly in the
    plot.

    Parameters
    ----------
    patient : Phantom
        Patient phantom from instance of class Phantom. Can be of phantom_model "plane",
        "cylinder" or "human"

    """
    if not isinstance(patient, Phantom):
        return 0.0

    if patient.phantom_model == PHANTOM_MODEL_PLANE:
        return -0.01

    return 0.0



import logging
from typing import List

import plotly.graph_objects as go

from ..beam_class import Beam
from ..constants import (
    COLOR_BEAM,
    COLOR_DETECTOR,
    COLOR_PAD,
    COLOR_PATIENT,
    COLOR_SOURCE,
    COLOR_TABLE,
    MESH_NAME_PAD,
    MESH_OPACITY_BEAM,
    PLOT_ASPECTMODE_SETUP_AND_EVENT,
    PLOT_AXIS_TITLE_X,
    PLOT_AXIS_TITLE_Y,
    PLOT_AXIS_TITLE_Z,
    PLOT_DRAGMODE,
    PLOT_FONT_FAMILY,
    PLOT_FONT_SIZE,
    PLOT_HOVERLABEL_FONT_FAMILY,
    PLOT_HOVERLABEL_FONT_SIZE,
    PLOT_LIGHTNING_AMBIENT,
    PLOT_LIGHTNING_DIFFUSE,
    PLOT_SOURCE_SIZE,
    PLOT_TITLE_FONT_FAMILY,
    PLOT_TITLE_FONT_SIZE,
    PLOT_WIREFRAME_LINE_WIDTH,
    PLOT_ZERO_LINE_WIDTH,
)
from ..phantom_class import Phantom
from .create_mesh3d import create_mesh_3d_general
from .create_plot_and_save_to_file import create_plot_and_save_to_file
from .create_wireframes import create_wireframes
from .get_camera_view import get_camera_view
from .plot_settings import fetch_plot_colors, fetch_plot_margin, fetch_plot_size

logger = logging.getLogger(__name__)


def create_setup_and_event_plot(
    patient: Phantom,
    table: Phantom,
    pad: Phantom,
    beam: Beam,
    mode: str,
    patient_text: List[str],
    source_text: List[str],
    table_text: List[str],
    detectors_text: List[str],
    pad_text: List[str],
    beam_text: List[str],
    title: str,
    dark_mode=True,
    notebook_mode: bool = False,
):

    logger.debug("Creating meshes for plot")

    COLOR_CANVAS, COLOR_PLOT_TEXT, COLOR_GRID, COLOR_ZERO_LINE = fetch_plot_colors(dark_mode=dark_mode)

    PLOT_WIDTH, PLOT_HEIGHT = fetch_plot_size(notebook_mode=notebook_mode)

    PLOT_MARGIN = fetch_plot_margin(notebook_mode=notebook_mode)

    patient_mesh = create_mesh_3d_general(
        obj=patient,
        color=COLOR_PATIENT,
        mesh_text=patient_text,
        lighting=dict(diffuse=PLOT_LIGHTNING_DIFFUSE, ambient=PLOT_LIGHTNING_AMBIENT),
    )

    source_mesh = go.Scatter3d(
        x=[beam.r[0, 0], beam.r[0, 0]],
        y=[beam.r[0, 1], beam.r[0, 1]],
        z=[beam.r[0, 2], beam.r[0, 2]],
        hoverinfo="text",
        mode="markers",
        marker=dict(size=PLOT_SOURCE_SIZE, color=COLOR_SOURCE),
        text=source_text,
    )

    table_mesh = create_mesh_3d_general(obj=table, color=COLOR_TABLE, mesh_text=table_text)

    detector_mesh = create_mesh_3d_general(obj=beam, color=COLOR_DETECTOR, mesh_text=detectors_text, detector_mesh=True)

    pad_mesh = create_mesh_3d_general(obj=pad, color=COLOR_PAD, mesh_text=pad_text, mesh_name=MESH_NAME_PAD)

    beam_mesh = create_mesh_3d_general(obj=beam, color=COLOR_BEAM, opacity=MESH_OPACITY_BEAM, mesh_text=beam_text)

    logger.debug("Create wireframes")
    wf_beam, wf_table, wf_pad, wf_detector = create_wireframes(
        beam=beam,
        table=table,
        pad=pad,
        line_width=PLOT_WIREFRAME_LINE_WIDTH,
        visible=True,
    )

    logger.debug("Setting up plot layout settings")
    layout = go.Layout(
        height=PLOT_HEIGHT,
        width=PLOT_WIDTH,
        margin=PLOT_MARGIN,
        font=dict(family=PLOT_FONT_FAMILY, size=PLOT_FONT_SIZE, color=COLOR_PLOT_TEXT),
        title=dict(
            font=dict(
                family=PLOT_TITLE_FONT_FAMILY,
                size=PLOT_TITLE_FONT_SIZE,
                color=COLOR_PLOT_TEXT,
            ),
            text=title,
        ),
        hoverlabel=dict(
            font=dict(
                family=PLOT_HOVERLABEL_FONT_FAMILY,
                size=PLOT_HOVERLABEL_FONT_SIZE,
                color=COLOR_PLOT_TEXT,
            )
        ),
        paper_bgcolor=COLOR_CANVAS,
        showlegend=False,
        dragmode=PLOT_DRAGMODE,
        scene=dict(
            aspectmode=PLOT_ASPECTMODE_SETUP_AND_EVENT,
            camera=get_camera_view(),
            xaxis=dict(
                title=PLOT_AXIS_TITLE_X,
                backgroundcolor=COLOR_CANVAS,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
            yaxis=dict(
                title=PLOT_AXIS_TITLE_Y,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                backgroundcolor=COLOR_CANVAS,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
            zaxis=dict(
                title=PLOT_AXIS_TITLE_Z,
                gridcolor=COLOR_GRID,
                linecolor=COLOR_GRID,
                backgroundcolor=COLOR_CANVAS,
                zerolinecolor=COLOR_ZERO_LINE,
                zerolinewidth=PLOT_ZERO_LINE_WIDTH,
            ),
        ),
    )

    data = [
        patient_mesh,
        source_mesh,
        table_mesh,
        detector_mesh,
        pad_mesh,
        beam_mesh,
        wf_beam,
        wf_table,
        wf_pad,
        wf_detector,
    ]

    create_plot_and_save_to_file(mode=mode, data=data, layout=layout)



from pyskindose.constants import KEY_PARAM_HUMAN_MESH, KEY_PARAM_PHANTOM_MODEL
from pyskindose.helpers.create_attributes_string import create_attributes_string
from pyskindose.settings.patient_offset import PatientOffset
from pyskindose.settings.phantom_dimensions import PhantomDimensions


class PhantomSettings:
    """A class for setting all the phantom related settings required.

    Attributes
    ----------
    model : str
        Select which model to represent the skin surface for skindose
        calculations. Valid selections: "plane" (2D planar surface),
        "cylinder" (cylinder with elliptical cross section) or "human" (phantom
        in the shape of a human, created with MakeHuman.)
    human_mesh: str
        Select which MakeHuman phantom to represent the patient when
        model = "human" is selected. Valid selections: Any of the .stl files
        in the folder phantom_data. Enter as a string without the .stl file
        ending.
    patient_offset : PhantomOffset
        Instance of class PhantomOffset containing patient - table isocenter
        offset.
    patient_orientation : str
        patient orientation on table. Choose between 'head_first_supine' and
        'feet_first_supine'.
    dimension : PhantomDimensions
        Instance of class PhantomDimensions containing all dimensions required
        to create any of the mathematical phantoms, which is all but human.

    """

    def __init__(self, ptm_dim: dict):
        """Initialize phantom settings class.

        Parameters
        ----------
        ptm_dim : PhantomDimensions
            Instance of class PhantomDimensions containing all dimensions for
            the mathematical phantoms. See attributes of PhantomDimensions.

        """
        self.model = ptm_dim[KEY_PARAM_PHANTOM_MODEL]
        self.human_mesh = ptm_dim[KEY_PARAM_HUMAN_MESH]
        self.patient_orientation = ptm_dim["patient_orientation"]
        self.patient_offset = PatientOffset(offset=ptm_dim["patient_offset"])
        self.dimension = PhantomDimensions(ptm_dim=ptm_dim["dimension"])

        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="phantom", indent_level=0)

    def update_attrs_str(self):
        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="phantom", indent_level=0)

    def to_printable_string(self, color: str = "light_slate_blue"):
        return (
            f"[bold {color}]Phantom settings[/bold {color}]\n"
            f"\t[{color}]model: {self.model}[/{color}]\n"
            f"\t[{color}]patient_offset:[/{color}]\n"
            f"\t\t[{color}]d_lat: {self.patient_offset.d_lat}[/{color}]\n"
            f"\t\t[{color}]d_ver: {self.patient_offset.d_ver}[/{color}]\n"
            f"\t\t[{color}]d_lon: {self.patient_offset.d_lon}[/{color}]\n"
            f"\t[{color}]patient_orientation: {self.patient_orientation}[/{color}]\n"
            f"\t[{color}]dimension:[/{color}]\n"
            f"\t\t[{color}]plane_length: {self.dimension.plane_length}[/{color}]\n"
            f"\t\t[{color}]plane_width: {self.dimension.plane_width}[/{color}]\n"
            f"\t\t[{color}]plane_resolution: {self.dimension.plane_resolution}[/{color}]\n"
            f"\t\t[{color}]cylinder_length: {self.dimension.cylinder_length}[/{color}]\n"
            f"\t\t[{color}]cylinder_radii_a: {self.dimension.cylinder_radii_a}[/{color}]\n"
            f"\t\t[{color}]cylinder_radii_b: {self.dimension.cylinder_radii_b}[/{color}]\n"
            f"\t\t[{color}]cylinder_resolution: {self.dimension.cylinder_resolution}[/{color}]\n"
            f"\t\t[{color}]table_thickness: {self.dimension.table_thickness}[/{color}]\n"
            f"\t\t[{color}]table_length: {self.dimension.table_length}[/{color}]\n"
            f"\t\t[{color}]table_width: {self.dimension.table_width}[/{color}]\n"
            f"\t\t[{color}]pad_thickness: {self.dimension.pad_thickness}[/{color}]\n"
            f"\t\t[{color}]pad_width: {self.dimension.pad_width}[/{color}]\n"
            f"\t\t[{color}]pad_length: {self.dimension.pad_length}[/{color}]\n"
        )



from .beam_class import Beam
from .geom_calc import position_patient_phantom_on_table, scale_field_area, fetch_and_append_hvl, check_new_geometry
from .rdsr_parser import rdsr_parser
from .rdsr_normalizer import rdsr_normalizer
from .phantom_class import Phantom
from pyskindose.plotting import plot_geometry
from .analyze_data import analyze_data
from .settings import PyskindoseSettings


def load_settings_example_json() -> dict:
    import json
    from pathlib import Path

    return json.loads((Path(__file__).parent / "settings_example.json").read_text())


def print_available_human_phantoms():
    from pathlib import Path

    phantom_data_dir = Path(__file__).parent / "phantom_data"
    phantoms = [
        phantom.stem for phantom in phantom_data_dir.glob("*.stl") if not phantom.stem.endswith("reduced_1000t")
    ]

    for phantom in phantoms:
        print(phantom)


def print_example_rdsr_files():
    rdsr_data_dir = get_path_to_example_rdsr_files()
    files = [file.name for file in rdsr_data_dir.glob("*.dcm")]

    print("Available RDSR files:\n")
    for filename in files:
        print(f"\t{filename}")

    print(f"\nFiles located in {rdsr_data_dir.absolute()}")


def get_path_to_example_rdsr_files():
    from pathlib import Path

    return Path(__file__).parent / "example_data" / "RDSR"



import logging
from typing import Any, Dict, List

import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline

from pyskindose import Phantom
from pyskindose import constants as c
from pyskindose.corrections import calculate_k_med

logger = logging.getLogger(__name__)


def add_corrections_and_event_dose_to_output(
    normalized_data: pd.DataFrame,
    event: int,
    hits: List[bool],
    table_hits: List[bool],
    patient: Phantom,
    back_scatter_interpolation: List[CubicSpline],
    field_area: List[float],
    k_tab: List[float],
    corrections_db: str,
    output: Dict[str, Any],
) -> Dict[str, Any]:
    """Add correction factors and event dose to output dictionary.

    Parameters
    ----------
    normalized_data : pd.DataFrame
        RDSR data, normalized for compliance with PySkinDose.
    event : int
        Irradiation event index.
    hits : List[bool]
        A boolean list of the same length as the number of patient skin cells. True for
        all entrance skin cells that are hit by the beam for a specific irradiation
        event.
    table_hits : List[bool]
        A boolean list that specfies (for each hit), if the bean passes through the
        patient support table, by default None
    patient : Phantom
        Patient phantom, either of type plane, cylinder or human, i.e. instance of class
        Phantom
    back_scatter_interpolation : List[CubicSpline]
        List of interpolation objects to used to estimate backscatter correction from
        the correction database
    field_area : List[float]
        X-ray field area in (cm^2) for each phantom skin cell that are hit by the X-ray
        beam
    k_tab : List[float]
        List of table correction factors
    corrections_db : str
        A string defining the path to the corrections SQLite db
    output : Dict[str, Any]
        Dictionary containing outputs to store from the calculations. E.g. dose map and
        correction factors.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing outputs to store from the calculations. E.g. dose map and
        correction factors.

    """
    event_dose = np.zeros(len(patient.r))
    if not sum(hits):
        output[c.OUTPUT_KEY_DOSE_MAP] += event_dose
        return output

    logger.debug("Calculating back scatter correction factor")
    k_bs = back_scatter_interpolation[event](np.sqrt(field_area))

    logger.debug("Calculating reference point medium correction (air -> water)")
    k_med = calculate_k_med(
        data_norm=normalized_data, field_area=field_area, event=event, corrections_db=corrections_db
    )

    output[c.OUTPUT_KEY_CORRECTION_BACK_SCATTER][event] = k_bs
    output[c.OUTPUT_KEY_CORRECTION_MEDIUM][event] = k_med
    output[c.OUTPUT_KEY_CORRECTION_TABLE][event] = k_tab[event]

    logger.debug("Calculating event skin dose by applying each correction" "factor to the reference point air kerma")

    event_dose[hits] += normalized_data.K_IRP[event]
    event_dose[hits] *= output[c.OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW][event]

    event_dose[hits] *= k_med
    event_dose[hits] *= k_bs

    temp = np.ones(len(table_hits))
    temp[table_hits] = k_tab[event]
    event_dose[hits] *= temp

    output[c.OUTPUT_KEY_DOSE_MAP] += event_dose

    return output



from pyskindose import constants as c

PARAMETERS = dict(
    # modes: 'calculate_dose', 'plot_setup', 'plot_event', 'plot_procedure'
    mode=c.MODE_PLOT_PROCEDURE,
    # RDSR filename
    rdsr_filename="S1.dcm",
    # Irrading event index for mode='plot_event'
    plot_event_index=12,
    # Set True to estimate table correction, or False to use measured k_tab
    estimate_k_tab=False,
    # Numeric value of estimated table correction
    k_tab_val=0.8,
    # plot settings
    plot={
        # dark mode for plots
        c.MODE_DARK_MODE: True,
        # notebook mode
        c.MODE_NOTEBOOK_MODE: False,
        # choose if dosemap should be plotted after dose calculations.
        c.MODE_PLOT_DOSEMAP: False,
        # colorscale for dosemaps
        c.DOSEMAP_COLORSCALE_KEY: c.DOSEMAP_COLORSCALE,
        # max number of event for inclusion of patient phantom in plot
        # procedure
        c.MAX_EVENT_FOR_PATIENT_INCLUSION_IN_PROCEDURE_KEY: 0,
        # Irrading event index for mode='plot_event'
        c.PLOT_EVENT_INDEX_KEY: 12,
    },
    # Phantom settings:
    phantom=dict(
        # Phantom model, valid selections: 'plane', 'cylinder', or 'human'
        model=c.PHANTOM_MODEL_HUMAN,
        # Human phantom .stl filename, without .stl ending.
        human_mesh=c.PHANTOM_MESH_ADULT_MALE,
        # Patient offset from table isocenter (centered at head end side).
        patient_offset={
            c.OFFSET_LONGITUDINAL_KEY: 0,
            c.OFFSET_VERTICAL_KEY: 0,
            c.OFFSET_LATERAL_KEY: -35,
        },
        # Dimensions of matematical phantoms (except model='human')
        patient_orientation=c.PATIENT_ORIENTATION_HEAD_FIRST_SUPINE,
        dimension={
            # Length of plane phantom
            c.DIMENSION_PLANE_LENGTH: 120,
            # Width of plane phantom
            c.DIMENSION_PLANE_WIDTH: 40,
            # Resolution of plane phantom
            c.DIMENSION_PLANE_RESOLUTION: c.RESOLUTION_SPARSE,
            # Length of cylinder phantom
            c.DIMENSION_CYLINDER_LENGTH: 150,
            # First radii of cylinder phantom
            c.DIMENSION_CYLINDER_RADII_A: 20,
            # Second radii of cylinder phantom
            c.DIMENSION_CYLINDER_RADII_B: 10,
            # Resolution of cylinder.
            c.DIMENSION_CYLINDER_RESOLUTION: c.RESOLUTION_DENSE,
            # Support table length
            c.DIMENSION_TABLE_LENGTH: 281.5,
            # Support table width
            c.DIMENSION_TABLE_WIDTH: 45,
            # Support table thickness
            c.DIMENSION_TABLE_THICKNESS: 5,
            # Support pad length
            c.DIMENSION_PAD_LENGTH: 281.5,
            # Support pad width
            c.DIMENSION_PAD_WIDTH: 45,
            # Support pad thickness
            c.DIMENSION_PAD_THICKNESS: 4,
        },
    ),
)



import logging
import os
import sqlite3

import pandas as pd

logger = logging.getLogger("pyskindose")


def db_connect(db_name: str):
    """Set up the database connection with tables needed for PSD calculations.

    Parameters
    ----------
    db_name : str, optional
        The name of/path to the sqlite3 database to connect to
        and if it doesn't exist, create, by default 'corrections.db'

    Returns
    -------
    conn
        connection to database
    cursor
        cursor to database connection

    """
    db_exist = False

    if os.path.exists(db_name):
        db_exist = True

    # Connect to database
    try:
        conn = sqlite3.connect(db_name)

        # Create cursor (enables sql commands using the sql method)
        cursor = conn.cursor()
    except sqlite3.Error as e:
        logger.error("Failed to connect to/create database at {}".format(db_name), exc_info=True)
        raise

    if not db_exist:

        # load data from CSV files

        # HVL table, simulated with SpekCalc.
        hvl_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "hvl_tables/hvl_combined.csv"))

        # Backscatter and mu_en/rho quotients, from Benmanhlouf et al.
        ks_table = pd.read_csv(
            os.path.join(os.path.dirname(__file__), "table_data", "correction_medium_and_backscatter.csv")
        )

        # Measured and approximated patient support table transmission.
        tab_pad_table = pd.read_csv(
            os.path.join(os.path.dirname(__file__), "table_data", "correction_table_and_pad_attenuation.csv")
        )

        # Table containing lab specific parameters.
        device_info_table = pd.read_csv(os.path.join(os.path.dirname(__file__), "table_data", "device_info.csv"))

        # Upload tables to database

        hvl_table.to_sql("hvl_combined", conn, if_exists="replace", index=False)

        ks_table.to_sql("correction_medium_and_backscatter", conn, if_exists="replace", index=False)

        tab_pad_table.to_sql("correction_table_and_pad_attenuation", conn, if_exists="replace", index=False)

        device_info_table.to_sql("device_info", conn, if_exists="replace", index=False)

        # Commits the current transactions
        conn.commit()

    return conn, cursor



from typing import Dict, Optional, Union


class RotationDirection:
    """Switch pos/neg direction of beam- and table angles.

    Attributes
    ----------
    keys of __init__ parameter directions
        Value of each attribute (i.e. key) is integers of either +1 or -1 to be
        used as a multiplicative correction factor to switch pos/neg direction
        in each angle. PySkinDose default angles are all +1.

    """

    def __init__(self, directions: Optional[Dict[str, str]] = None):
        """Initialize class attributes.

        Parameters
        ----------
        directions : dict
            dictionary with keys 'Ap1', 'Ap2', 'Ap3', 'At1', 'At2' and 'At3'.
            Each key contains either '+' or '-'.

        """
        self.Ap1: Optional[int] = None if directions is None else self._get_direction_as_value(directions["Ap1"])
        self.Ap2: Optional[int] = None if directions is None else self._get_direction_as_value(directions["Ap2"])
        self.Ap3: Optional[int] = None if directions is None else self._get_direction_as_value(directions["Ap3"])
        self.At1: Optional[int] = None if directions is None else self._get_direction_as_value(directions["At1"])
        self.At2: Optional[int] = None if directions is None else self._get_direction_as_value(directions["At2"])
        self.At3: Optional[int] = None if directions is None else self._get_direction_as_value(directions["At3"])

    def update_rotation_direction(self, directions: Dict[str, str]):
        self.Ap1 = (
            self._get_direction_as_value(directions["Ap1"])
            if self.Ap1 is None
            else self._get_direction_as_value(self.Ap1)
        )
        self.Ap2 = (
            self._get_direction_as_value(directions["Ap2"])
            if self.Ap2 is None
            else self._get_direction_as_value(self.Ap2)
        )
        self.Ap3 = (
            self._get_direction_as_value(directions["Ap3"])
            if self.Ap3 is None
            else self._get_direction_as_value(self.Ap3)
        )
        self.At1 = (
            self._get_direction_as_value(directions["At1"])
            if self.At1 is None
            else self._get_direction_as_value(self.At1)
        )
        self.At2 = (
            self._get_direction_as_value(directions["At2"])
            if self.At2 is None
            else self._get_direction_as_value(self.At2)
        )
        self.At3 = (
            self._get_direction_as_value(directions["At3"])
            if self.At3 is None
            else self._get_direction_as_value(self.At3)
        )

    @staticmethod
    def _get_direction_as_value(direction: Union[str, int]):

        if direction in (-1, +1):
            return direction

        if direction in ("+", "-"):
            return 1 if direction == "+" else -1

        raise ValueError(f"The direction must be given as '+', '-', 1, or -1, but was given as {direction}")



COLOR_CANVAS_DARK = "rgb(33,33,33)"
COLOR_CANVAS_LIGHT = "rgb(252, 252, 252)"
COLOR_PLOT_TEXT_LIGHT = "rgb(52, 49, 49)"
COLOR_PLOT_TEXT_DARK = "rgb(252, 252, 252)"
COLOR_ZERO_LINE_LIGHT = "green"
COLOR_ZERO_LINE_DARK = "limegreen"
COLOR_BEAM = "red"
COLOR_DETECTOR = "#D3D3D3"
COLOR_PAD = "slateblue"
COLOR_PATIENT = "#CE967C"
COLOR_SOURCE = "#D3D3D3"
COLOR_TABLE = "#D3D3D3"
COLOR_SLIDER_BACKGROUND = "red"
COLOR_SLIDER_BORDER_LIGHT = "rgb(252,252,252)"
COLOR_SLIDER_BORDER_DARK = "rgb(45,45,45)"
COLOR_SLIDER_TICK_LIGHT = "rgb(52, 49, 49)"
COLOR_SLIDER_TICK_DARK = "white"

COLOR_GRID_DARK = "rgb(99,99,99)"
COLOR_GRID_LIGHT = "rgb(99,99,99)"

COLOR_WIRE_FRAME_BEAM = "red"
COLOR_WIRE_FRAME_DETECTOR = "darkslategray"
COLOR_WIRE_FRAME_PAD = "#3f3f3f"
COLOR_WIRE_FRAME_TABLE = "#3f3f3f"

DATA_DS_IRP = "DSIRP"

KEY_PARAM_MODE = "mode"
KEY_PARAM_RDSR_FILENAME = "rdsr_filename"
KEY_PARAM_ESTIMATE_K_TAB = "estimate_k_tab"
KEY_PARAM_K_TAB_VAL = "k_tab_val"
KEY_PARAM_PHANTOM_MODEL = "model"
KEY_PARAM_HUMAN_MESH = "human_mesh"
KEY_PARAM_INHERENT_FILTRATION = "inherent_filtration"
KEY_PARAM_REMOVE_INVALID_ROWS = "remove_invalid_rows"
KEY_PARAM_SILENCE_PYDICOM_WARNINGS = "silence_pydicom_warnings"

DIMENSION_PLANE_LENGTH = "plane_length"
DIMENSION_PLANE_RESOLUTION = "plane_resolution"
DIMENSION_PLANE_WIDTH = "plane_width"

DIMENSION_CYLINDER_LENGTH = "cylinder_length"
DIMENSION_CYLINDER_RADII_A = "cylinder_radii_a"
DIMENSION_CYLINDER_RADII_B = "cylinder_radii_b"
DIMENSION_CYLINDER_RESOLUTION = "cylinder_resolution"

DIMENSION_TABLE_LENGTH = "table_length"
DIMENSION_TABLE_WIDTH = "table_width"
DIMENSION_TABLE_THICKNESS = "table_thickness"

DIMENSION_PAD_LENGTH = "pad_length"
DIMENSION_PAD_WIDTH = "pad_width"
DIMENSION_PAD_THICKNESS = "pad_thickness"

DIMENSION_UNIT_KEY = "unit"
DIMENSION_UNIT_CM = "cm"

FIELD_SIZE_MODE_ACTUAL_SHUTTER_DISTANCE = "ASD"
FIELD_SIZE_MODE_COLLIMATED_FIELD_AREA = "CFA"

KEY_RDSR_ACQUISITION_DATA = "AcquisitionData"
KEY_RDSR_COMMENT = "Comment"
KEY_RDSR_CONCEPT_CODE_SEQUENCE = "ConceptCodeSequence"
KEY_RDSR_CONTENT_SEQUENCE = "ContentSequence"
KEY_RDSR_DATE_TIME = "DateTime"
KEY_RDSR_DETECTORSIZE_MM = "DetectorSize_mm"
KEY_RDSR_EVENT_XRAY_DATA = "Irradiation Event X-Ray Data"
KEY_RDSR_II_DIAMETER_SRDATA = "iiDiameter SRData"
KEY_RDSR_MANUFACTURER = "Manufacturer"
KEY_RDSR_MANUFACTURER_MODEL_NAME = "ManufacturerModelName"
KEY_RDSR_MEASURED_VALUE_SEQUENCE = "MeasuredValueSequence"
KEY_RDSR_TEXT_VALUE = "TextValue"
KEY_RDSR_UID = "UID"
KEY_RDSR_FILTER_MATERIAL_COPPER = "Copper or Copper compound"
KEY_RDSR_FILTER_MATERIAL_ALUMINUM = "Aluminum or Aluminum compound"
KEY_RDSR_FILTER_MIN = "XRayFilterThicknessMinimum_mm"
KEY_RDSR_FILTER_MAX = "XRayFilterThicknessMaximum_mm"
KEY_RDSR_FILTER_MATERIAL = "XRayFilterMaterial"
KEY_RDSR_FILTER_TYPE = "XRayFilterType"
KEY_RDSR_DISTANCE_SOURCE_DETECTOR = "DistanceSourcetoDetector_mm"

KEY_NORMALIZATION_DETECTOR_SIDE_LENGTH = "detector_side_length"
KEY_NORMALIZATION_FIELD_SIZE_MODE = "field_size_mode"
KEY_NORMALIZATION_KVP = "kVp"
KEY_NORMALIZATION_MANUFACTURER = "manufacturer"
KEY_NORMALIZATION_MODELS = "models"
KEY_NORMALIZATION_FILTER_SIZE_ALUMINUM = "filter_thickness_Al"
KEY_NORMALIZATION_FILTER_SIZE_COPPER = "filter_thickness_Cu"
KEY_NORMALIZATION_DISTANCE_SOURCE_DETECTOR = "DSD"
KEY_NORMALIZATION_DISTANCE_SOURCE_ISOCENTER = "DSI"
KEY_NORMALIZATION_DISTANCE_ISOCENTER_DETECTOR = "DID"
KEY_NORMALIZATION_DISTANCE_SOURCE_IRP = "DSIRP"
KEY_NORMALIZATION_MODEL_NAME = "model"
KEY_NORMALIZATION_ACQUISITION_TYPE = "acquisition_type"
KEY_NORMALIZATION_ACQUISITION_PLANE = "acquisition_plane"
KEY_NORMALIZATION_AIR_KERMA = "K_IRP"


IRRADIATION_EVENT_PROCEDURE_KEY_BEAM = "Beam"
IRRADIATION_EVENT_PROCEDURE_KEY_DETECTORS = "Detectors"
IRRADIATION_EVENT_PROCEDURE_KEY_PAD = "Pad"
IRRADIATION_EVENT_PROCEDURE_KEY_PATIENT = "Patient"
IRRADIATION_EVENT_PROCEDURE_KEY_SOURCE = "Source"
IRRADIATION_EVENT_PROCEDURE_KEY_TABLE = "Table"
IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_BEAM = "WF Beam"
IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_DETECTORS = "WF Detectors"
IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_PAD = "WF Pad"
IRRADIATION_EVENT_PROCEDURE_KEY_WIRE_FRAME_TABLE = "WF Table"
IRRADIATION_EVENT_STEP_KEY_ARGUMENTS = "args"
IRRADIATION_EVENT_STEP_KEY_LABEL = "label"
IRRADIATION_EVENT_STEP_KEY_METHOD = "method"

MAX_EVENT_FOR_PATIENT_INCLUSION_IN_PROCEDURE_KEY = "max_events_for_patient_inclusion"

MESH_NAME_PAD = "Support pad"
MESH_OPACITY_BEAM = 0.4

MODE_CALCULATE_DOSE = "calculate_dose"
MODE_PLOT_EVENT = "plot_event"
MODE_PLOT_PROCEDURE = "plot_procedure"
MODE_PLOT_SETUP = "plot_setup"
MODE_PLOT_DOSEMAP = "plot_dosemap"
MODE_DARK_MODE = "dark_mode"
MODE_INTERACTIVITY = "interactivity"

OUTPUT_KEY_CORRECTION_BACK_SCATTER = "k_bs"
OUTPUT_KEY_CORRECTION_INVERSE_SQUARE_LAW = "k_isq"
OUTPUT_KEY_CORRECTION_MEDIUM = "k_med"
OUTPUT_KEY_CORRECTION_TABLE = "k_tab"
OUTPUT_KEY_DOSE_MAP = "dose_map"
OUTPUT_KEY_HITS = "hits"
OUTPUT_KEY_KERMA = "kerma"

PHANTOM_MESH_ADULT_MALE = "hudfrid"

PHANTOM_MODEL_CYLINDER = "cylinder"
PHANTOM_MODEL_HUMAN = "human"
PHANTOM_MODEL_PAD = "pad"
PHANTOM_MODEL_PLANE = "plane"
PHANTOM_HUMAN_MESH_SPARSE_MODEL_ENDING = "_reduced_1000t"
PHANTOM_MODEL_TABLE = "table"


PLOT_AXIS_TITLE_X = "X - LON [cm]"
PLOT_AXIS_TITLE_Y = "Y - VER [cm]"
PLOT_AXIS_TITLE_Z = "Z - LAT [cm]"

PLOT_FONT_FAMILY = "Arial"
PLOT_FONT_SIZE = 14

PLOT_HOVERLABEL_FONT_FAMILY = "Arial"
PLOT_HOVERLABEL_FONT_SIZE = 14

PLOT_TITLE_FONT_FAMILY = "Arial"
PLOT_TITLE_FONT_SIZE = 14

PLOT_EVENT_INDEX_KEY = "plot_event_index"

PLOT_SLIDER_BORDER_WIDTH = 3
PLOT_SLIDER_FONT_SIZE_CURRENT = 18
PLOT_SLIDER_FONT_SIZE_GENERAL = 14
PLOT_SLIDER_PADDING = dict(b=0, t=0, l=25, r=25)
PLOT_SLIDER_PADDING_NOTEBOOK = dict(b=0, t=0, l=20, r=20)

PLOT_WIREFRAME_LINE_WIDTH = 4
PLOT_DRAGMODE = "orbit"

PLOT_ASPECTMODE_SETUP_AND_EVENT = "data"
PLOT_ASPECTMODE_PLOT_DOSEMAP = "data"

PLOT_ASPECTMODE_PLOT_PROCEDURE = "cube"
PLOT_ORDER_STATIC = ["right", "back", "left", "front"]

PLOT_PROCEDURE_AXIS_RANGE_X = [-300, 300]
PLOT_PROCEDURE_AXIS_RANGE_Y = [-300, 300]
PLOT_PROCEDURE_AXIS_RANGE_Z = [-300, 300]

# shift size of the entire image to the left to remove whitespace in plot
PLOT_GROUND_SHIFT_X_STATIC = 50

# shift size of each of the four subplots to hide all but 1 colorscale
PLOT_SHIFT_X_STATIC = [0 * 70, 1 * 70, 2 * 70, 3 * 70]

# size of each static subplot
PLOT_BASE_SIZE_STATIC = 297


PLOT_SOURCE_SIZE = 8
PLOT_FILE_TYPE_STATIC = ".png"

PLOT_LIGHTNING_DIFFUSE = 0.5
PLOT_LIGHTNING_AMBIENT = 0.5

PLOT_SLIDER_TRANSITION = dict(duration=300, easing="quad-in-out")

PLOT_ZERO_LINE_WIDTH = 5

# PLOT_HEIGHT_NOTEBOOK = 800
# PLOT_WIDTH_NOTEBOOK = None
PLOT_HEIGHT_NOTEBOOK = 900
PLOT_WIDTH_NOTEBOOK = 600


PLOT_HEIGHT = None
PLOT_WIDTH = None

PLOT_MARGIN_NOTEBOOK = dict(l=5, r=5, b=5, t=40)
PLOT_MARGIN = dict(l=0, r=0, b=100, t=40)

# plot camera angles for static dosemaps
PLOT_EYE_BACK = dict(eye=dict(x=0, y=+2.5, z=0))
PLOT_EYE_FRONT = dict(eye=dict(x=0, y=-2.5, z=0))
PLOT_EYE_LEFT = dict(eye=dict(x=2.5, y=+1.5, z=0))
PLOT_EYE_RIGHT = dict(eye=dict(x=-2.5, y=+1.5, z=0))

PLOT_TRACE_ORDER_PHANTOM_WIREFRAME = [0, 4, 5, 1, 0, 3, 7, 4, 7, 6, 5, 1, 2, 3, 2, 6]
PLOT_TRACE_ORDER_BEAM_WIREFRAME = [0, 1, 2, 0, 2, 3, 0, 3, 4, 0, 1, 4, 0]
PLOT_TRACE_ORDER_DETECTOR_WIREFRAME = [0, 0 + 4, 0, 1, 1 + 4, 1, 2, 2 + 4, 2, 3, 3 + 4, 3, 0, 4, 5, 6, 7, 4]

OFFSET_LATERAL_KEY = "d_lat"
OFFSET_VERTICAL_KEY = "d_ver"
OFFSET_LONGITUDINAL_KEY = "d_lon"

RESOLUTION_SPARSE = "sparse"
RESOLUTION_DENSE = "dense"

RUN_ARGUMENTS_MODE_HEADLESS = "headless"
RUN_ARGUMENTS_MODE_GUI = "gui"
RUN_ARGUMENTS_OUTPUT_DICT = "dict"
RUN_ARGUMENTS_OUTPUT_JSON = "json"
RUN_ARGUMENTS_OUTPUT_HTML = "html"
RUN_ARGUMENTS_VALID_OUTPUT_FORMATS = (RUN_ARGUMENTS_OUTPUT_DICT, RUN_ARGUMENTS_OUTPUT_HTML, RUN_ARGUMENTS_OUTPUT_JSON)

TABULATED_FIELD_SIDE_LENGTHS_IN_CM = [5, 10, 20, 25, 35]

# Offset needed to show plane phantom correctly
VISUAL_OFFSET_PHANTOM_MODEL_PLANE = -0.01
MODE_NOTEBOOK_MODE = "notebook_mode"

PLOT_HEIGHT_KEY = "height"
PLOT_WIDTH_KEY = "width"

DOSEMAP_COLORSCALE_KEY = "colorscale"
# Available colorscales are documented here:
# https://plotly.com/python/builtin-colorscales/
DOSEMAP_COLORSCALE = "jet"
# amp

PATIENT_ORIENTATION_HEAD_FIRST_SUPINE = "head_first_supine"
PATIENT_ORIENTATION_FEET_FIRST_SUPINE = "feet_first_supine"



from pyskindose.helpers.create_attributes_string import create_attributes_string


class PhantomDimensions:
    """A class for setting the phantom dimensions for mathematical phantoms.

    Attributes
    ----------
    plane_length : int
        Lenth of plane phantom.
    plane_width : int
        Width of plane phantom.
    plane_resolution: str
        Select either 'sparse' or 'dense' resolution of the skin cell grid
        on the surface of the plane phantom. Note: dense is more computational
        expensive.
    cylinder_length : int
        Length of cylider phantom.
    cylinder_radii_a : float
        First radii of the cylindrical cross section of the cylindrical
        phantom, which lies in the "width" direction.
    cylinder_radii_b : float
        Second radii of the cylindrical cross section of the cylindrical
        phantom, which lies in the "thickness" direction. radii a should
        be greater than radii b.
    cylinder_resolution: str
        Select either 'sparse' or 'dense' resolution of the skin cell grid
        on the surface of the elliptical cylinder. Note: dense is more
        computational expensive.
    table_thickness : int
        Thickness of the support table phantom.
    table_length : int
        Length of the support table phantom.
    table_width : int
        Width of the support table phantom.
    pad_thickness : int
        Thickness of the patient support table phantom.
    pad_width : int
        Width of the patient support table phantom.
    pad_length : int
        Length of the patient support table phantom.

    """

    def __init__(self, ptm_dim: dict):
        """Initialize phantom dimension class.

        Parameters
        ----------
        ptm_dim : dict
            Dictionary containing all of the phantom dimensions that are
            appended as attributes to this class, see class attributes.

        """
        self.plane_length = ptm_dim["plane_length"]
        self.plane_width = ptm_dim["plane_width"]
        self.plane_resolution = ptm_dim["plane_resolution"]
        self.cylinder_length = ptm_dim["cylinder_length"]
        self.cylinder_radii_a = ptm_dim["cylinder_radii_a"]
        self.cylinder_radii_b = ptm_dim["cylinder_radii_b"]
        self.cylinder_resolution = ptm_dim["cylinder_resolution"]
        self.table_thickness = ptm_dim["table_thickness"]
        self.table_length = ptm_dim["table_length"]
        self.table_width = ptm_dim["table_width"]
        self.pad_thickness = ptm_dim["pad_thickness"]
        self.pad_width = ptm_dim["pad_width"]
        self.pad_length = ptm_dim["pad_length"]

        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="dimensions", indent_level=1)

    def update_attrs_str(self):
        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="dimension", indent_level=1)

    def to_dict_pad(self):
        return {
            "plane_length": self.plane_length,
            "plane_width": self.plane_width,
            "plane_resolution": self.plane_resolution,
            "cylinder_radii_a": self.cylinder_radii_a,
            "cylinder_length": self.cylinder_length,
            "cylinder_radii_b": self.cylinder_radii_b,
            "cylinder_resolution": self.cylinder_resolution,
            "table_thickness": self.table_thickness,
            "table_length": self.table_length,
            "table_width": self.table_width,
            "pad_thickness": self.pad_thickness,
            "pad_width": self.pad_width,
            "pad_length": self.pad_length,
        }

    def to_dict_cylinder(self):
        return {
            "cylinder_length": self.cylinder_length,
            "cylinder_radii_a": self.cylinder_radii_a,
            "cylinder_radii_b": self.cylinder_radii_b,
            "cylinder_resolution": self.cylinder_resolution,
        }



from typing import Any, Dict, List, Optional

import pandas as pd

from pyskindose.constants import (
    KEY_NORMALIZATION_DETECTOR_SIDE_LENGTH,
    KEY_NORMALIZATION_FIELD_SIZE_MODE,
    KEY_NORMALIZATION_MANUFACTURER,
    KEY_NORMALIZATION_MODELS,
    KEY_RDSR_MANUFACTURER,
    KEY_RDSR_MANUFACTURER_MODEL_NAME,
)
from pyskindose.helpers.create_attributes_string import create_attributes_string
from pyskindose.settings.rotation_direction import RotationDirection
from pyskindose.settings.translation_direction import TranslationDirection
from pyskindose.settings.translation_offset import TranslationOffset


class NormalizationSettings:
    """A class to normalize RDSR for PySkinDose compliance.

    Attributes
    ----------
    trans_offset : pyskindose.settings.translation_offset.TranslationOffset
        See class variables of _TranslationOffset
    trans_dir : pyskindose.settings.translation_direction.TranslationDirection
        See class variables of _TranslationDirection
    rot_dir : pyskindose.settings.rotation_direction.RotationDirection
        See class variables of _RotationDirection
    field_size_mode : str
        method for calculating field size at image receptor plane.
        Choose either "CFA" (collimated field area) or "ACD" (actual shutter
        distance). For more info, see calculate_field_size in geom_calc.py.
    detector_side_length : str
        side length of active image receptor area in cm.

    """

    def __init__(self, normalization_settings: List[Dict[str, Any]]):
        """Initialize class attributes."""
        self.normalization_settings_list: List[Dict[str, Any]] = normalization_settings
        self.trans_offset: TranslationOffset = TranslationOffset()
        self.trans_dir: TranslationDirection = TranslationDirection()
        self.rot_dir: RotationDirection = RotationDirection()
        self.field_size_mode: Optional[str] = None
        self.detector_side_length: Optional[str] = None

        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="normalization", indent_level=0)

    def update_used_settings(self, data_parsed: pd.DataFrame):
        manufacturer = data_parsed[KEY_RDSR_MANUFACTURER][0].casefold()
        model = data_parsed[KEY_RDSR_MANUFACTURER_MODEL_NAME][0].casefold()

        setting = [
            setting
            for setting in self.normalization_settings_list
            if manufacturer == setting[KEY_NORMALIZATION_MANUFACTURER].casefold()
            and model in [mod.casefold() for mod in setting[KEY_NORMALIZATION_MODELS]]
        ]

        if not setting:
            raise NotImplementedError(
                f"Could not find settings for the given manufacturer and model ({manufacturer=}, {model=})"
            )

        setting = setting[0]

        self.trans_offset.update_translation_offset(offset=setting.get("translation_offset"))
        self.trans_dir.update_translation_direction(directions=setting.get("translation_direction"))
        self.rot_dir.update_rotation_direction(directions=setting.get("rotation_direction"))
        self.field_size_mode = setting[KEY_NORMALIZATION_FIELD_SIZE_MODE]
        self.detector_side_length = setting[KEY_NORMALIZATION_DETECTOR_SIDE_LENGTH]

        self.update_attrs_str()

    def update_attrs_str(self):
        self.attrs_str = create_attributes_string(attrs_parent=self, object_name="normalization", indent_level=0)

    def to_printable_string(self, color: str = "blue"):
        return (
            f"[bold {color}]Normalization settings[/bold {color}]\n"
            f"\t[{color}]trans_offset:[/{color}]\n"
            f"\t\t[{color}]x:{self.trans_offset.x}[/{color}]\n"
            f"\t\t[{color}]y:{self.trans_offset.y}[/{color}]\n"
            f"\t\t[{color}]z:{self.trans_offset.z}[/{color}]\n"
            f"\t[{color}]trans_dir:[/{color}]\n"
            f"\t\t[{color}]x:{self.trans_dir.x}[/{color}]\n"
            f"\t\t[{color}]y:{self.trans_dir.y}[/{color}]\n"
            f"\t\t[{color}]z:{self.trans_dir.z}[/{color}]\n"
            f"\t[{color}]rot_dir:[/{color}]\n"
            f"\t\t[{color}]Ap1:{self.rot_dir.Ap1}[/{color}]\n"
            f"\t\t[{color}]Ap2:{self.rot_dir.Ap2}[/{color}]\n"
            f"\t\t[{color}]Ap3:{self.rot_dir.Ap3}[/{color}]\n"
            f"\t\t[{color}]At1:{self.rot_dir.At1}[/{color}]\n"
            f"\t\t[{color}]At2:{self.rot_dir.At2}[/{color}]\n"
            f"\t\t[{color}]At3:{self.rot_dir.At3}[/{color}]\n"
            f"\t[{color}]field_size_mode: {self.field_size_mode}[/{color}]\n"
            f"\t[{color}]detector_side_length: {self.detector_side_length}[/{color}]\n"
        )



import numpy as np
import pandas as pd


def calculate_rotation_matrices(normalized_data: pd.DataFrame) -> pd.DataFrame:
    at = normalized_data.loc[:, ["At1", "At2", "At3"]]
    at.At1 = [np.deg2rad(val) for val in at["At1"]]  # Table Horizontal Rotation Angle, rotation axis: center of table
    at.At2 = [np.deg2rad(val) for val in at["At2"]]  # Table Head Tilt Angle, rotation axis: center of table
    at.At3 = [np.deg2rad(val) for val in at["At3"]]  # Table Cradle Tilt Angle

    return normalized_data.join(
        pd.DataFrame(
            [
                (
                    [
                        [+1, +0, +0],
                        [+0, +float(np.cos(row.At2)), -float(np.sin(row.At2))],
                        [+0, +float(np.sin(row.At2)), +float(np.cos(row.At2))],
                    ],
                    [
                        [+float(np.cos(row.At1)), +0, +float(np.sin(row.At1))],
                        [+0, +1, +0],
                        [-float(np.sin(row.At1)), +0, +float(np.cos(row.At1))],
                    ],
                    [
                        [+float(np.cos(row.At3)), -float(np.sin(row.At3)), +0],
                        [+float(np.sin(row.At3)), +float(np.cos(row.At3)), +0],
                        [+0, +0, +1],
                    ],
                )
                for row in at.itertuples()
            ],
            columns=["Rx", "Ry", "Rz"],
        )
    )
